"""
The RPC(Remote Procedure Call) module.

Handles GET, SET and PUT communication with the server using JSON encoded
messages.

"""
import json
import datetime
import os
import logging
from . import message_data
from jsonrpcclient import requests, response

JSONRPC = '2.0'


class SeluxitRpc:
    """
    Check connection status.

    Invokes an init_ok() method on reference of connection. If error with
    communication occurs then a connection is closed.

    Args:
        connection: A reference to the socket instance.
        error_msg: Error message while one occurs.

    """

    @staticmethod
    def create_meta(network, network_id):
        """Create meta.

        Creates and returns a meta data about a network.

        Args:
            network:
            network_id: Identify number of a network.

        Returns:
            meta data dictionary.

        """
        return {
            'id': network_id,
            'type': '{}'.format(network),
            'version': '2.0'
        }

    @staticmethod
    def get_rpc_success_response(message_id):
        """Get success response.

        Returns a success, JSON formatted, encoded in utf-8 response.

        Args:
            message_id: Id of a message.

        Returns:
            JSON formatted data.

        """
        success_response = str(response.SuccessResponse(jsonrpc=JSONRPC,
                                                        id=message_id,
                                                        result=True))
        return json.loads(success_response)

    @staticmethod
    def get_rpc_fail_response(message_id, text):
        """
        Get fail response.

        Returns a fail, JSON formatted, encoded in utf-8 response with a error
        message.

        Args:
            message_id: Id of a message.
            text: A message of error description.

        Returns:
            JSON formatted data.

        """
        error_description = {'message': text, 'code': -32020, 'data': ''}
        error_response = str(response.ErrorResponse(jsonrpc=JSONRPC,
                                                    id=message_id,
                                                    error=error_description))
        return json.loads(error_response)

    @staticmethod
    def is_upgradable():
        """
        Check if System is set to upgradable.

        Returns:
            True, if the system is upgradable,
            False, if it is not.

        """
        return os.getenv("UPGRADABLE", "").upper() == "TRUE"

    def __init__(self):
        """
        Initialize the seluxit_rpc class.

        Initializes an object of seluxit_rpc class by passing required
        parameters. While initialization, wapp_log is created.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

    def get_rpc_network(self, network_id, network_name, verb,
                        trace_id=None):
        """
        Retrieve network from server.

        Retrieves network from a server. The method is used while adding a
        network. Depending on request method it fills data_json_rpc dictionary
        with different information.

        Args:
            network_id: Unique identifying number of a network
            network_name: Name of a network
            verb: indicates what verb should be used.
            trace_id:  ID of the debug trace {default: None}

        Returns:
            JSON formatted data of network

        """
        meta = self.create_meta('network', network_id)
        if SeluxitRpc.is_upgradable():
            meta.update({'upgradable': True})
        data_inside = {
            'meta': meta,
            'name': network_name
        }
        url = '/network'

        if verb == message_data.PUT:
            url = '/{}'.format(network_id)

        if trace_id:
            url = "{}?trace={}".format(url, trace_id)

        data_json_rpc = requests.Request(verb,
                                         url=url,
                                         data=data_inside)

        return data_json_rpc

    def get_rpc_state(
            self,
            data,
            network_id,
            device_id,
            value_id,
            state_id,
            set_type,
            verb,
            trace_id=None
    ):
        """
        Retrieve state of the value.

        Retrieves the value state from a server. Uses functionality of
        base_device_value to get a directory with common attributes of value
        class.

        Args:
            data: Data from the state of the value.
            network_id: Unique identifier of the network.
            device_id: Unique identifier of a device.
            value_id: Unique identifier of a value.
            state_id: Unique identifier of the state.
            set_type: The type to set.
            verb: indicates what verb should be used.
            trace_id: ID of the debug trace. (default: {None})

        Returns:
            JSON formatted data of the state.

        """
        update = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        device_state = {
            'meta': self.create_meta('state', state_id),
            'type': set_type,
            'status': 'Send',
            'data': data,
            'timestamp': update
        }

        url = '/network/{}/device/{}/value/{}/state'
        url = url.format(network_id, device_id, value_id)

        if verb == message_data.GET:
            device_state = None
        url = '{}/{}'.format(url, state_id)

        if trace_id:
            url = '{}?trace={}'.format(url, trace_id)

        data_json_rpc = requests.Request(verb,
                                         url=url,
                                         data=device_state)
        return data_json_rpc

    def get_rpc_delete(self,
                       network_id,
                       device_id,
                       value_id,
                       state_id,
                       trace_id=None):
        """
        Creates delete request.

        The method is used to create message that could allow to
        delete network or its elements.

        Args:
            network_id: id of the network to delete/modify.
            device_id: id of the device to delete/modify.
            value_id: id of the value to delete/modify.
            state_id: id of the state to delete.
            trace_id: ID of the debug trace. {default: None}

        Returns:
            JSON formatted data of delete message

        """
        if network_id:
            url = '/network/{}'.format(network_id)
            if device_id:
                url += '/device/{}'.format(device_id)
                if value_id:
                    url += '/value/{}'.format(value_id)
                    if state_id:
                        url += '/state/{}'.format(state_id)

        if trace_id:
            url = '{}?trace={}'.format(url, trace_id)

        data_json_rpc = requests.Request('DELETE',
                                         url=url)

        return data_json_rpc

    def get_rpc_whole_json(self, json_data, trace_id=None):
        """
        Creates request containing the whole json file.

        The method is used while starting initializing and it fills
        data_json_rpc dictionary with url and data containing json object.

        Args:
            json_data: Data read from json file.
            trace_id:  ID of the debug trace {default: None}

        Returns:
            JSON formatted data of network

        """
        url = '/{}'.format("network")

        if trace_id:
            url = "{}?trace={}".format(url, trace_id)

        data_json_rpc = requests.Request('POST',
                                         url=url,
                                         data=json_data)
        return data_json_rpc



"""
Custom Exception module.

This module contains custom exceptions used throughout the wappsto module.
"""


class DeviceNotFoundException(Exception):
    """
    Exception to signify Device not being found.

    This custom Exception extends the Exception class and implements no
    custom methods.

    Attributes:
        pass: The Exception has no custom body,
        just extends the Exception superclass.

    """

    pass


class ValueNotFoundException(Exception):
    """
    Exception to signify a value not being found.

    This custom Exception extends the Exception class and implements no
    custom methods.

    Attributes:
        pass: The Exception has no custom body,
        just extends the Exception superclass.

    """

    pass


class ServerConnectionException(Exception):
    """
    Exception to signify a server connection issue.

    This custom Exception extends the Exception class and implements no
    custom methods.

    Attributes:
        pass: The Exception has no custom body,
        just extends the Exception superclass.

    """

    pass


class CallbackNotCallableException(Exception):
    """
    Exception to signify Callback not being callable.

    This custom Exception extends the Exception class and implements no
    custom methods.

    Attributes:
        pass: The Exception has no custom body,
        just extends the Exception superclass.

    """

    pass



"""
The value module.

Stores attributes for the value instance and handles value-related
methods.
"""
import logging
import warnings
import datetime
import decimal
import threading
from ..connection import message_data
from ..errors import wappsto_errors


class Value:
    """
    Value instance.

    Stores attributes for the value instance and handles value-related
    methods.
    """

    def __init__(
        self,
        parent,
        uuid,
        name,
        type_of_value,
        data_type,
        permission,
        number_max,
        number_min,
        number_step,
        number_unit,
        string_encoding,
        string_max,
        blob_encoding,
        blob_max,
        period,
        delta
    ):
        """
        Initialize the Value class.

        Initializes an object of value class by passing required parameters.

        Args:
            parent: Reference to a device object
            uuid: An unique identifier of a device
            name: A name of a device
            type_of_value: Determines a type of value [e.g temperature, CO2]
            data_type: Defines whether a value is string, blob or number
            permission: Defines permission [read, write, read and write]
            (if data_type is number then these parameters are relevant):
            number_max: Maximum number a value can have
            number_min: Minimum number a value can have
            number_step: Number defining a step
            number_unit: Unit in which a value should be read
            (if data_type is string then these parameters are relevant):
            string_encoding: A string encoding of a value
            string_max: Maximum length of string
            (if data_type is blob then these parameters are relevant):
            blob_encoding: A blob encoding of a value
            blob_max: Maximum length of a blob

            period: defines the time after which a value should send report
                message. Default: {None})
            delta: defines the a difference of value (default: {None})

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.parent = parent
        self.uuid = uuid
        self.name = name
        self.type_of_value = type_of_value
        self.data_type = data_type
        self.permission = permission
        # The value shared between state instances.
        self.number_max = number_max
        self.number_min = number_min
        self.number_step = number_step
        self.number_unit = number_unit
        self.string_encoding = string_encoding
        self.string_max = string_max
        self.blob_encoding = blob_encoding
        self.blob_max = blob_max
        self.report_state = None
        self.control_state = None
        self.callback = None

        self.timer = threading.Timer(None, None)
        self.last_update_of_report = None

        if period:
            self.set_period(period)
        if delta:
            self.set_delta(delta)

        msg = "Value {} debug: {}".format(name, str(self.__dict__))
        self.wapp_log.debug(msg)

    def __getattr__(self, attr):  # pragma: no cover
        """
        Get attribute value.

        When trying to get value from last_controlled warning is raised about
        it being deprecated and calls get_data instead.

        Returns:
            value of get_data

        """
        if attr in ["last_controlled"]:
            warnings.warn("Property %s is deprecated" % attr)
            return self.get_control_state().data

    def set_period(self, period):
        """
        Set the value reporting period.

        Sets the time defined in second to report a value to
        the server and starts timer.

        Args:
            period: Reporting period.

        """
        if period is None:
            self.wapp_log.warning("Period value is not provided.")
            return

        try:
            period = int(period)
        except ValueError:
            self.wapp_log.error("Period value must be a number.")
            return

        if period <= 0:
            self.wapp_log.warning("Period value must be greater then 0.")
            return

        if self.get_report_state() is not None:
            self.period = period
            self.__set_timer()
            self.wapp_log.debug("Period successfully set.")
        else:
            self.wapp_log.warning("Cannot set the period for this value.")

    def __set_timer(self):
        """
        Set timer.

        Stop previous timer and sets new one if period value is not None.

        """
        self.timer.cancel()
        if self.period is not None:
            self.timer = threading.Timer(self.period, self.__timer_done)
            self.timer.start()

    def __timer_done(self):
        self.__set_timer()
        self.timer_elapsed = True
        self.handle_refresh()

    def set_delta(self, delta):
        """
        Set the delta to report between.

        Sets the delta (range) of change to report in. When a change happens
        in the range of this delta it will be reported.

        Args:
            delta: Range to report between.

        """
        if delta is None:
            self.wapp_log.warning("Delta value is not provided.")
            return

        try:
            delta = float(delta)
        except ValueError:
            self.wapp_log.error("Delta value must be a number")
            return

        if delta <= 0:
            self.wapp_log.warning("Delta value must be greater then 0.")
            return

        if self.__is_number_type() and self.get_report_state():
            self.delta = delta
            self.wapp_log.debug("Delta successfully set.")
        else:
            self.wapp_log.warning("Cannot set the delta for this value.")

    def get_parent_device(self):  # pragma: no cover
        """
        Retrieve parent device reference.

        Gets a reference to the device that owns this device.

        Returns:
            Reference to instance of Device class that owns this Value.

        """
        return self.parent

    def add_report_state(self, state):
        """
        Set report state reference to the value list.

        Adds a report state reference to the Value class.

        Args:
            state: Reference to instance of State class.

        """
        self.report_state = state
        msg = "Report state {} has been added.".format(state)
        self.wapp_log.debug(msg)

    def add_control_state(self, state):
        """
        Set control state reference to the value list.

        Adds a control state reference to the Value class.

        Args:
            state: Reference to instance of State class.

        """
        self.control_state = state
        msg = "Control state {} has been added".format(state)
        self.wapp_log.debug(msg)

    def get_report_state(self):
        """
        Retrieve child report state reference.

        Gets a reference to the child State class.

        Returns:
            Reference to instance of State class.

        """
        if self.report_state is not None:
            return self.report_state
        else:
            msg = "Value {} has no report state.".format(self.name)
            self.wapp_log.warning(msg)

    def get_control_state(self):
        """
        Retrieve child control state reference.

        Gets a reference to the child State class.

        Returns:
            Reference to instance of State class.

        """
        if self.control_state is not None:
            return self.control_state
        else:
            msg = "Value {}  has no control state.".format(self.name)
            self.wapp_log.warning(msg)

    def get_now():
        """
        Retrieve current time.

        Using datetime library returns current time.

        Returns:
            Current time in format [%Y-%m-%dT%H:%M:%S.%fZ].

        """
        return datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')

    def set_callback(self, callback):
        """
        Set the callback.

        Sets the callback attribute.

        Args:
            callback: Callback reference.

        Raises:
            CallbackNotCallableException: Custom exception to signify invalid
            callback.

        """
        if not callable(callback):
            msg = "Callback method should be a method"
            self.wapp_log.error("Error setting callback: {}".format(msg))
            raise wappsto_errors.CallbackNotCallableException
        self.callback = callback
        self.wapp_log.debug("Callback {} has been set.".format(callback))
        return True

    def __validate_value_data(self, data_value):
        if self.__is_number_type():
            try:
                data_value = self.ensure_number_value_follows_steps(data_value)

                if data_value is None:
                    return None

                if self.number_min <= data_value <= self.number_max:
                    return str(data_value)
                else:
                    msg = "Invalid number. Range: {}-{}. Your: {}".format(
                        self.number_min,
                        self.number_max,
                        str(data_value)
                    )
                    self.wapp_log.warning(msg)
            except ValueError:
                msg = "Invalid type of value. Must be a number: {}".format(str(data_value))
                self.wapp_log.error(msg)
        elif self.__is_string_type():
            if (self.string_max is None
                    or len(str(data_value)) <= int(self.string_max)):
                return data_value
            else:
                msg = ("Value {} not in correct range for {}"
                       .format(data_value, self.name))
                self.wapp_log.warning(msg)
        elif self.__is_blob_type():
            if (self.blob_max is None
                    or len(str(data_value)) <= int(self.blob_max)):
                return data_value
            else:
                msg = ("Value {} not in correct range for {}"
                       .format(data_value, self.name))
                self.wapp_log.warning(msg)
        else:
            msg = "Value type {} is invalid".format(self.date_type)
            self.wapp_log.error(msg)

    def ensure_number_value_follows_steps(self, data_value):
        """
        Ensure number value follows steps.

        Converts values to decimal and ensures number step is always positive,
        ensures that data value follows steps and removes exes 0's after
        decimal point.

        Args:
            data_value: float value indicating current state of value.

        Returns:
            data_value

        """
        try:
            data_value = decimal.Decimal(str(data_value))
            number_step = abs(decimal.Decimal(str(self.number_step)))

            result = data_value % number_step
            if result < 0:
                result += number_step
            data_value = data_value - result

            data_value = '{:f}'.format(data_value)
            data_value = (data_value.rstrip('0').rstrip('.')
                          if '.' in data_value else data_value)

            return decimal.Decimal(data_value)
        except decimal.InvalidOperation as e:
            self.wapp_log.error("Invalid operation: {}".format(e))

    def update(self, data_value, timestamp=get_now()):
        """
        Update value.

        Check if value has a state and validates the information in data_value
        if both of these checks pass then method send_state is called.

        Args:
            data_value: the new value.
            timestamp: time of action.

        Returns:
            True/False indicating the result of operation.

        """
        if not self.check_delta_and_period(data_value):
            return False

        state = self.get_report_state()
        if state is None:
            self.wapp_log.warning("Value is write only.")
            return False

        data_value = self.__validate_value_data(data_value)
        if data_value is None:
            return False

        state.timestamp = timestamp

        msg = message_data.MessageData(
            message_data.SEND_REPORT,
            data=data_value,
            network_id=state.parent.parent.parent.uuid,
            device_id=state.parent.parent.uuid,
            value_id=state.parent.uuid,
            state_id=state.uuid,
            verb=message_data.PUT
        )
        self.parent.parent.conn.send_data.send_report(msg)

    def check_delta_and_period(self, data_value):
        """
        Check if delta and period allows data to be sent.

        Check if value has delta or period, if it has then if it passes
        checks then True is returned, otherwise False is returned.

        Args:
            data_value: the new value.

        Returns:
            True/False indicating the result of operation.

        """
        if (self.delta is not None and self.__is_number_type()):
            # delta should work
            data_value = float(data_value)
            if (self.last_update_of_report is None or abs(data_value - self.last_update_of_report) >= self.delta):
                # delta exeeded
                self.last_update_of_report = data_value
                if self.period is not None:
                    # timer should be reset if period exists
                    self.__set_timer()
                return True
            else:
                # delta not exeeded
                return self.check_period(False)
        return self.check_period(True)

    def check_period(self, return_value):
        """
        Check if period allows data to be sent.

        Check if value has period, if it has then if it passes
        checks then True is returned, otherwise False is returned.

        Args:
            return_value: default return value.

        Returns:
            True/False indicating the result of operation.

        """
        if self.period is not None:
            # period should work
            if self.timer_elapsed:
                # timer has elapsed
                self.timer_elapsed = False
                return True
            else:
                # timer is working
                return False
        return return_value

    def get_data(self):
        """
        Get value from report state.

        Check if value has a report state if it has return its data else return
        None.

        Returns:
            Value of report state.

        """
        state = self.get_report_state()
        if state is None:
            return None
        else:
            return state.data

    def handle_refresh(self):
        """
        Handles the refresh request.

        Calls __call_callback method with input of 'refresh'

        Returns:
            results of __call_callback

        """
        self.__call_callback('refresh')

    def handle_delete(self):
        """
        Handle delete.

        Calls the __call_callback method with initial input of "remove".

        Returns:
            result of __call_callback method.

        """
        self.__call_callback('remove')

    def delete(self):
        """
        Delete this object.

        Sends delete request for this object and removes its reference
        from parent.

        """
        message = message_data.MessageData(
            message_data.SEND_DELETE,
            network_id=self.parent.parent.uuid,
            device_id=self.parent.uuid,
            value_id=self.uuid
        )
        self.parent.parent.conn.sending_queue.put(message)
        self.parent.values.remove(self)
        self.wapp_log.info("Value removed")

    def __call_callback(self, event):
        if self.callback is not None:
            self.callback(self, event)

    def handle_control(self, data_value):
        """
        Handles the control request.

        Sets data value of control_state object, with value provided and calls
        __call_callback method with input of 'set'.

        Args:
            data_value: the new value.

        Returns:
            results of __call_callback

        """
        self.control_state.data = data_value

        return self.__call_callback('set')

    def __is_number_type(self):
        """
        Validate data type.

        Checks whether the type of the value is number.

        Returns:
        True if the type is number otherwise false.
        boolean

        """
        return self.data_type == "number"

    def __is_string_type(self):
        """
        Validate data type.

        Checks whether the type of the value is string.

        Returns:
        True if the type is string otherwise false.
        boolean

        """
        return self.data_type == "string"

    def __is_blob_type(self):
        """
        Validate data type.

        Checks whether the type of the value is blob.

        Returns:
        True if the type is blob otherwise false.
        boolean

        """
        return self.data_type == "blob"



"""Server connection package.

The package handles connecting to the server, as well as all the related
issues. It also starts a sending and receiving thread with the resoective
sending/receiving queue.
"""



"""
The network module.

Stores attributes for the network instance.
"""
import logging
from ..connection import message_data
from ..errors import wappsto_errors


class Network:
    """
    Network instance class.

    Stores attributes for the network instance.
    """

    def __init__(self, uuid, version, name, devices, data_manager):
        """
        Initialize the Network class.

        Initializes an object of network class by passing required parameters.

        Args:
            uuid: Unique identifier of a network
            version: Version of a network
            name: Name of a network
            devices: list of devices in network
            data_manager: Instance of DataManager

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.uuid = uuid
        self.version = version
        self.name = name
        self.devices = devices
        self.data_manager = data_manager
        self.rpc = None
        self.conn = None
        self.callback = None
        msg = "Network {} Debug \n{}".format(name, str(self.__dict__))
        self.wapp_log.debug(msg)

    def set_callback(self, callback):
        """
        Set the callback.

        Sets the callback attribute.

        Args:
            callback: Callback reference.

        Raises:
            CallbackNotCallableException: Custom exception to signify invalid
            callback.

        """
        if not callable(callback):
            msg = "Callback method should be a method"
            self.wapp_log.error("Error setting callback: {}".format(msg))
            raise wappsto_errors.CallbackNotCallableException
        self.callback = callback
        self.wapp_log.debug("Callback {} has been set.".format(callback))
        return True

    def get_device(self, name):
        """
        Device reference.

        Finds the device with a specific name attribute based on a given
        string

        Args:
            name: String containing the name attribute to search for.

        Returns:
            A reference to the device object instance.

        Raises:
            DeviceNotFoundException: Device {name} not found.

        """
        for device in self.devices:
            if name == device.name:
                return device
        else:
            msg = "Device {} not found".format(name)
            raise wappsto_errors.DeviceNotFoundException(msg)

    def handle_delete(self):
        """
        Handle delete.

        Calls the __call_callback method with initial input of "remove".

        Returns:
            result of __call_callback method.

        """
        self.__call_callback('remove')

    def delete(self):
        """
        Delete this object.

        Sends delete request for this object and removes its reference
        from parent.

        """
        message = message_data.MessageData(
            message_data.SEND_DELETE,
            network_id=self.uuid,
        )
        self.conn.sending_queue.put(message)
        self.data_manager.network = None
        self.wapp_log.info("Network removed")

    def __call_callback(self, event):
        if self.callback is not None:
            self.callback(self, event)



from setuptools import find_packages, setup

import codecs
import os
import re


with open("README.md", "r") as file:
    long_description = file.read()


here = os.path.abspath(os.path.dirname(__file__))


def find_version(*file_paths):
    path = os.path.join(here, *file_paths)
    with codecs.open(path, 'r') as fp:
        version_file = fp.read()
        version_match = re.search(r"__version__ = ['\"]([^'\"]*)['\"]",
                                  version_file, re.M)
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")


setup(
    name="wappsto",
    version=find_version("wappsto", "__init__.py"),
    author="Seluxit A/S",
    author_email="support@seluxit.com",
    license="Apache-2.0",
    description="Python Package to connect to wappsto.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wappsto/wappsto-python",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    packages=find_packages(),
    tests_require=[
        'pytest',
    ],
    install_requires=[
       'jsonrpcclient'
    ],
    python_requires='>3.4.0',
)



"""
The __init__ method for the wappsto module.

Stores the Wappsto class functionality.
"""

import inspect
import logging
import os
import signal

from threading import Event

from .connection import seluxit_rpc
from .connection import communication
from .errors import wappsto_errors
from .connection import event_storage
from . import status
from .data_operation import data_manager

RETRY_LIMIT = 5


class Object_instantiation:
    # For backward compability.
    def __init__(self):
        self.status = status


object_instantiation = Object_instantiation()


class Wappsto:
    """
    The main package classs.

    Establishes a connection to the wappsto server, initializes the classes,
    starts a sending/receiving thread.
    """

    __version__ = "1.2.3"

    def __init__(self, json_file_name=None, load_from_state_file=False,
                 log_offline=False, log_location="logs",
                 log_data_limit=10, limit_action=event_storage.REMOVE_OLD,
                 compression_period=event_storage.DAY_PERIOD):
        # TODO(Dimitar): Come up with a better description.
        """
        Initialize wappsto class.

        Creates a wappsto object which methods are exposed for users to use.
        These are: starting and stopping connection with a server and
        retrieving devices. These methods are enough for users to make
        operations on their network.

        Args:
            json_file_name: name of a json file containing all information
                about a network (default: {None})
            load_from_state_file: Defines if the data should be loaded from
                saved files (default: {False})
            log_offline: boolean indicating if data should be logged (default: {False})
            log_location: location of the logs (default: {"logs"})
            log_data_limit: limit of data to be saved in log [in Megabytes] (default: {10})
            limit_action: action to take when limit is reached (default: {REMOVE_OLD})
            compression_period: period for compressing data [day, hour] (default: {DAY_PERIOD})

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

        # TODO(Dimitar): Comment on this later.
        stack = inspect.stack()[1][1]
        self.path_to_calling_file = os.path.dirname(os.path.abspath(stack))

        self.connecting = True
        self.rpc = seluxit_rpc.SeluxitRpc()
        self.event_storage = event_storage.OfflineEventStorage(
            log_offline,
            log_location,
            log_data_limit,
            limit_action,
            compression_period
        )
        self.socket = None
        self.receive_thread = None
        self.send_thread = None
        self.status = status.Status()
        try:
            self.data_manager = data_manager.DataManager(
                json_file_name=json_file_name,
                load_from_state_file=load_from_state_file,
                path_to_calling_file=self.path_to_calling_file
            )
        except FileNotFoundError as fnfe:
            self.wapp_log.error("Failed to open file: {}".format(fnfe))
            self.stop(False)
            raise fnfe

    def get_status(self):
        """
        Wappsto status.

        Retrieves the status class instance which is set to different statuses
        throughout the program execution.

        Returns:
            A reference to the status object instance.

        """
        return self.status

    def get_network(self):
        """
        Wappsto network.

        Retrieves the network class instance.

        Returns:
            A reference to the network object instance.

        """
        return self.data_manager.network

    def get_devices(self):
        """
        Wappsto devices.

        Retrieves the devices list containing instances of devices

        Returns:
            A list of devices.

        """
        return self.data_manager.network.devices

    def get_by_id(self, id):
        """
        Wappsto get by id.

        Retrieves the instance of a class if its id matches the provided one

        Args:
            id: unique identifier used for searching

        Returns:
            A reference to the network/device/value/state object instance.

        """
        return self.data_manager.get_by_id(id)

    def set_status_callback(self, callback):
        """
        Sets callback for status.

        Sets the provided callback to the instance of status.

        Args:
            callback: reference to callback

        """
        self.status.set_callback(callback)

    def set_network_callback(self, callback):
        """
        Sets callback for network.

        Sets the provided callback to the instance of network.

        Args:
            callback: reference to callback

        """
        self.data_manager.network.set_callback(callback)

    def set_value_callback(self, callback):
        """
        Sets callback for values.

        Sets the provided callback to all instances of value.

        Args:
            callback: reference to callback

        """
        for device in self.data_manager.network.devices:
            for value in device.values:
                value.set_callback(callback)

    def get_device(self, name):
        """
        Device reference.

        Finds the device with a specific name attribute based on a given
        string

        Args:
            name: String containing the name attribute to search for.

        Returns:
            A reference to the device object instance.

        """
        for device in self.data_manager.network.devices:
            if name == device.name:
                return device
        else:
            msg = "Device {} not found in {}".format(name, self.data_manager)
            self.wapp_log.warning(msg, exc_info=True)
            self.stop(False)
            raise wappsto_errors.DeviceNotFoundException(msg)

    def start(self, address="wappsto.com", port=11006, automatic_trace=False, blocking=False):
        """
        Start the server connection.

        Creates a socket object instance and passes the address and port to it.

        Args:
            address: Address to connect the service to.
                (default: {"wappsto.com"})
            port: Port to connect the address to. (default: {11006})
            automatic_trace: indicates if all messages automaticaly send trace.
            blocking: Wheather or not this call should be a block call.
                      (default: False)
                      If sat to True, it will listen for a SIGTERM or SIGINT,
                      and terminate if those was received.

        """
        self.status.set_status(status.STARTING)

        self.socket = communication.ClientSocket(
            rpc=self.rpc,
            data_manager=self.data_manager,
            address=address,
            port=port,
            path_to_calling_file=self.path_to_calling_file,
            wappsto_status=self.status,
            automatic_trace=automatic_trace,
            event_storage=self.event_storage
        )

        self.status.set_status(status.CONNECTING)
        try:
            if not self.socket.connect():
                self.socket.reconnect(RETRY_LIMIT, send_reconnect=False)
        except wappsto_errors.ServerConnectionException as ce:
            self.stop(False)
            raise ce

        self.status.set_status(status.INITIALIZING)
        # Initializes the network, and all the subsequent devices, values and
        # states.
        # TODO(Dimitar): Change from generic Exception
        try:
            self.socket.initialize_all()
        except Exception as e:
            self.wapp_log.error("Error initializing: {}".format(e))
            self.stop(False)
            raise e

        self.status.set_status(status.STARTING_THREADS)
        # Starts the sending & receiving threads.
        try:
            self.receive_thread = self.socket.receiving_thread.start()
            self.send_thread = self.socket.sending_thread.start()
        except Exception as e:
            msg = "Error starting threads: {}".format(e)
            self.wapp_log.error(msg, exc_info=True)
            self.stop(False)
            raise e

        self.status.set_status(status.RUNNING)

        if blocking:
            self.keep_running()

    def keep_running(self):
        """
        Keeps wappsto running.

        Waiting for a SIGTERM or SIGINT, or self.terminated to be set,
        before it will exits.

        """
        self.terminated = Event()

        signal.signal(signal.SIGINT, lambda *args: self.terminated.set())
        signal.signal(signal.SIGTERM, lambda *args: self.terminated.set())
        self.wapp_log.info("Waiting terminate request.")

        self.terminated.wait()
        self.wapp_log.info("Terminate request received.")
        self.stop()

    def stop(self, save=True):
        """
        Stop the Wappsto service.

        Closes the connection socket to the server and, if the boolean is set
        to True it saves an instance of the current state.

        Args:
            save: Flag to determine whether runtime instances should be saved.
                (default: {True})

        """
        self.connecting = False
        self.status.set_status(status.DISCONNECTING)
        # Closes the socket connection, if one is established.
        if self.socket:
            self.socket.close()
        if save:
            self.data_manager.save_instance()
        self.wapp_log.info("Exiting...")
